# Copyright 2022 The KCL Authors. All rights reserved.

import os
import sys
import platform
import typing
import json
import inspect

from ctypes import *

import google.protobuf.json_format as json_format

import kclvm.compiler.extension.plugin.plugin as kcl_plugin
import kclvm.kcl.info as kcl_info
import kclvm.internal.gpyrpc.gpyrpc_pb2 as pb2
import kclvm.api.object as objpkg
import kclvm.kcl.error as kcl_error
import kclvm.config


kclvm_PANIC_INFO_KEY = "__kcl_PanicInfo__"

# Using kclvm rust cli PATH or current exec path.
_exe_root = os.path.dirname(os.environ.get("KCLVM_CLI_BIN_PATH") or sys.executable)
_cli_dll = None


def init_cli_dll():
    global _cli_dll

    if _cli_dll:
        return

    if platform.system() == "Darwin":
        _cli_dll_path = f"{_exe_root}/bin/libkclvm_cli_cdylib.dylib"
        _cli_dll = CDLL(_cli_dll_path)
    elif platform.system() == "Linux":
        _cli_dll_path = f"{_exe_root}/bin/libkclvm_cli_cdylib.so"
        _cli_dll = CDLL(_cli_dll_path)
    elif platform.system() == "Windows":
        _cli_dll_path = f"{_exe_root}/kclvm_cli_cdylib.dll"
        _cli_dll = CDLL(_cli_dll_path)
    else:
        raise f"unknown os: {platform.system()}"


class PluginContex:
    def __init__(self):
        self._plugin_dict: typing.Dict[str, any] = {}

    def call_method(self, name: str, args_json: str, kwargs_json: str) -> str:
        return self._call_py_method(name, args_json, kwargs_json)

    def _call_py_method(self, name: str, args_json: str, kwargs_json: str) -> str:
        try:
            return self._call_py_method_unsafe(name, args_json, kwargs_json)
        except Exception as e:
            return json.dumps({"__kcl_PanicInfo__": f"{e}"})

    def _get_plugin(self, plugin_name: str) -> typing.Optional[any]:
        if plugin_name in self._plugin_dict:
            return self._plugin_dict[plugin_name]

        module = kcl_plugin.get_plugin(plugin_name)
        self._plugin_dict[plugin_name] = module
        return module

    def _call_py_method_unsafe(
        self, name: str, args_json: str, kwargs_json: str
    ) -> str:
        dotIdx = name.rfind(".")
        if dotIdx < 0:
            return ""

        modulePath = name[:dotIdx]
        mathodName = name[dotIdx + 1 :]

        plugin_name = modulePath[modulePath.rfind(".") + 1 :]

        module = self._get_plugin(plugin_name)
        mathodFunc = None

        for func_name, func in inspect.getmembers(module):
            if func_name == kcl_info.demangle(mathodName):
                mathodFunc = func
                break

        args = []
        kwargs = {}

        if args_json:
            args = json.loads(args_json)
            if not isinstance(args, list):
                return ""
        if kwargs_json:
            kwargs = json.loads(kwargs_json)
            if not isinstance(kwargs, dict):
                return ""

        result = mathodFunc(*args, **kwargs)
        return json.dumps(result)


__plugin_context__ = PluginContex()
__plugin_method_agent_buffer__ = create_string_buffer(1024)


@CFUNCTYPE(c_char_p, c_char_p, c_char_p, c_char_p)
def plugin_method_agent(method: str, args_json: str, kwargs_json: str) -> c_char_p:
    method = str(method, encoding="utf-8")
    args_json = str(args_json, encoding="utf-8")
    kwargs_json = str(kwargs_json, encoding="utf-8")

    json_result = __plugin_context__.call_method(method, args_json, kwargs_json)

    global __plugin_method_agent_buffer__
    __plugin_method_agent_buffer__ = create_string_buffer(json_result.encode("utf-8"))
    return addressof(__plugin_method_agent_buffer__)


def kclvm_cli_run(args: pb2.ExecProgram_Args) -> str:
    init_cli_dll()

    _cli_dll.kclvm_cli_run.restype = c_char_p
    _cli_dll.kclvm_cli_run.argtypes = [c_char_p, c_void_p]

    args_json = json_format.MessageToJson(
        args, including_default_value_fields=True, preserving_proto_field_name=True
    )

    result_json = _cli_dll.kclvm_cli_run(args_json.encode("utf-8"), plugin_method_agent)
    return result_json.decode(encoding="utf-8")


def kclvm_cli_native_run_dylib(args: pb2.ExecProgram_Args) -> objpkg.KCLResult:
    json_result = kclvm_cli_run(args)
    warn_json_result = ""

    if json_result.startswith("ERROR:"):
        warn_json_result = json_result[len("ERROR:") :]
        json_result = "{}"

    try:
        data = json.loads(json_result)
    except Exception as e:
        raise Exception(f"Exception={e}, json_result={json_result}")

    panic_info = {}
    if kclvm_PANIC_INFO_KEY in data:
        panic_info = data
    else:
        if warn_json_result:
            try:
                panic_info = json.loads(warn_json_result)
            except Exception as e:
                raise Exception(f"Exception={e}, warn_json_result={warn_json_result}")
        else:
            panic_info = {}

    # check panic_info
    if panic_info.get(kclvm_PANIC_INFO_KEY):
        err_type_code = panic_info["err_type_code"]
        if err_type_code:
            err_type = kcl_error.ErrType((err_type_code,))
        else:
            err_type = kcl_error.ErrType.EvaluationError_TYPE

        file_msg = [
            kcl_error.ErrFileMsg(
                filename=panic_info.get("kcl_file"),
                line_no=panic_info.get("kcl_line"),
                col_no=panic_info.get("kcl_col"),
                arg_msg=panic_info.get("kcl_arg_msg"),
            )
        ]
        if kclvm.config.debug and kclvm.config.verbose >= 2:
            rust_filename = panic_info.get("rust_file")
            rust_line = panic_info.get("rust_line")
            rust_col = panic_info.get("rust_col")
            print(f"Rust error info: {rust_filename}:{rust_line}:{rust_col}")

        config_meta_file_msg = kcl_error.ErrFileMsg(
            filename=panic_info.get("kcl_config_meta_file"),
            line_no=panic_info.get("kcl_config_meta_line"),
            col_no=panic_info.get("kcl_config_meta_col"),
            arg_msg=panic_info.get("kcl_config_meta_arg_msg"),
        )
        if config_meta_file_msg.arg_msg:
            file_msg.append(config_meta_file_msg)

        if panic_info.get("is_warning") or panic_info.get("is_warnning"):
            kcl_error.report_warning(
                err_type=err_type, file_msgs=[], arg_msg=panic_info.get("message")
            )
        else:
            kcl_error.report_exception(
                err_type=err_type,
                file_msgs=file_msg,
                arg_msg=panic_info.get("message"),
            )

    return objpkg.KCLResult(data, os.path.abspath(args.k_filename_list[-1]))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gpyrpc/protorpc_wire.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1agpyrpc/protorpc_wire.proto\x12\rprotorpc_wire\"}\n\rRequestHeader\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x17\n\x0fraw_request_len\x18\x03 \x01(\r\x12%\n\x1dsnappy_compressed_request_len\x18\x04 \x01(\r\x12\x10\n\x08\x63hecksum\x18\x05 \x01(\r\"\x7f\n\x0eResponseHeader\x12\n\n\x02id\x18\x01 \x01(\x04\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x18\n\x10raw_response_len\x18\x03 \x01(\r\x12&\n\x1esnappy_compressed_response_len\x18\x04 \x01(\r\x12\x10\n\x08\x63hecksum\x18\x05 \x01(\r*.\n\x05\x43onst\x12\x08\n\x04ZERO\x10\x00\x12\x1b\n\x16MAX_REQUEST_HEADER_LEN\x10\x80\x08\x62\x06proto3')

_CONST = DESCRIPTOR.enum_types_by_name['Const']
Const = enum_type_wrapper.EnumTypeWrapper(_CONST)
ZERO = 0
MAX_REQUEST_HEADER_LEN = 1024


_REQUESTHEADER = DESCRIPTOR.message_types_by_name['RequestHeader']
_RESPONSEHEADER = DESCRIPTOR.message_types_by_name['ResponseHeader']
RequestHeader = _reflection.GeneratedProtocolMessageType('RequestHeader', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTHEADER,
  '__module__' : 'gpyrpc.protorpc_wire_pb2'
  # @@protoc_insertion_point(class_scope:protorpc_wire.RequestHeader)
  })
_sym_db.RegisterMessage(RequestHeader)

ResponseHeader = _reflection.GeneratedProtocolMessageType('ResponseHeader', (_message.Message,), {
  'DESCRIPTOR' : _RESPONSEHEADER,
  '__module__' : 'gpyrpc.protorpc_wire_pb2'
  # @@protoc_insertion_point(class_scope:protorpc_wire.ResponseHeader)
  })
_sym_db.RegisterMessage(ResponseHeader)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CONST._serialized_start=301
  _CONST._serialized_end=347
  _REQUESTHEADER._serialized_start=45
  _REQUESTHEADER._serialized_end=170
  _RESPONSEHEADER._serialized_start=172
  _RESPONSEHEADER._serialized_end=299
# @@protoc_insertion_point(module_scope)



# Copyright 2021 The KCL Authors. All rights reserved.

import os
import typing
import sys
import subprocess
import glob
import platform
import json
import inspect

import kclvm.config
import kclvm.kcl.ast as ast
import kclvm.kcl.info as kcl_info
import kclvm.api.object as objpkg
import kclvm.compiler.build.compiler as compiler
import kclvm.compiler.extension.plugin.plugin as kcl_plugin
import kclvm.kcl.error as kcl_error

from .native_runner import (
    LL_FILE_PATTERN,
    KCLVM_PANIC_INFO_KEY,
)


def native_run_wasm32(
    path_list: typing.List[str], *, ast_prog: ast.Program
) -> objpkg.KCLResult:
    from kclvm.internal.kclx.transformer import transform_ast_to_kclx_ast_json_str

    # Config

    if platform.system() == "Windows":
        _no_link = True
        _executable_root = os.path.dirname(sys.executable)
        _kclvm_cli = f"{_executable_root}\\kclvm_cli.exe"
        _clang = f"{_executable_root}\\tools\\clang\\bin\\clang.exe"
        _clang = _clang if os.path.exists(_clang) else "clang.exe"

        _kclvm_bc = f"{_executable_root}\\include\\_kclvm.bc"
        _kclvm_lib_path = f"{_executable_root}\\libs"
        _kclvm_lib_name = "kclvm_wasm32"
        _kclvm_undefined_file = f"{_executable_root}\\libs\\_kclvm_undefined_wasm.txt"

    else:
        _no_link = True
        _executable_root = os.path.dirname(os.path.dirname(sys.executable))
        _kclvm_cli = f"{_executable_root}/bin/kclvm_cli"
        _clang = f"{_executable_root}/tools/clang/bin/clang"
        _clang = _clang if os.path.exists(_clang) else "clang"

        _kclvm_bc = f"{_executable_root}/include/_kclvm.bc"
        _kclvm_lib_path = f"{_executable_root}/lib"
        _kclvm_lib_name = "kclvm_wasm32"
        _kclvm_undefined_file = f"{_executable_root}/lib/_kclvm_undefined_wasm.txt"

    _a_out_ast_json = "_a.out.ast.json"
    _a_out_ll = "_a.out.ll"
    _a_out_wasm = "_a.out.wasm"
    _out_bc_files = []

    # Resolve Program
    compiler.CompileProgram(ast_prog)

    # Build Program with kclvm_cli, windows donot support cache

    if True:
        # Transfrom Program
        ast_json = transform_ast_to_kclx_ast_json_str(ast_prog)
        with open(_a_out_ast_json, "w") as file:
            file.write(ast_json)
        if _no_link and os.path.exists(_a_out_ll):
            os.remove(_a_out_ll)
        if _no_link:
            _out_bc_files = glob.glob(_a_out_ll + LL_FILE_PATTERN)
            for file in _out_bc_files:
                if os.path.exists(file):
                    os.remove(file)

        # kclvm compile
        try:
            args = [
                _kclvm_cli,
                "build",
                _a_out_ast_json,
                "--bc",
                _kclvm_bc,
                "-o",
                _a_out_ll + ".tmp.ll",
            ]
            subprocess.check_call(args)

            fix_ll_local_name(_a_out_ll, _a_out_ll + ".tmp.ll")
            os.remove(_a_out_ll + ".tmp.ll")

        except subprocess.CalledProcessError as e:
            raise e
        _out_bc_files = glob.glob(_a_out_ll + LL_FILE_PATTERN)

        # clang
        try:
            args = [
                _clang,
                "--target=wasm32-unknown-unknown-wasm",
                "-Wno-override-module",
                "-nostdlib",
                "-Wl,--no-entry",
                "-Wl,--export-all",
                f"-Wl,--allow-undefined-file={_kclvm_undefined_file}",
                "-O3",
                f"-L{_kclvm_lib_path}",
                f"-l{_kclvm_lib_name}",
                _a_out_ll,
                "-o",
                _a_out_wasm,
            ]

            if platform.system() == "Windows":
                subprocess.check_call(args, stdout=open(os.devnull, "wb"))
            else:
                subprocess.check_call(args)

        except subprocess.CalledProcessError as e:
            raise e

    # run wasm
    result = WasmApp(path_list, _a_out_wasm).run()
    if not kclvm.config.debug:
        if os.path.exists(_a_out_ast_json):
            os.remove(_a_out_ast_json)
        if os.path.exists(_a_out_ll):
            os.remove(_a_out_ll)
        if _no_link:
            for file in _out_bc_files:
                if os.path.exists(file):
                    os.remove(file)
        if os.path.exists(_a_out_wasm):
            os.remove(_a_out_wasm)

    return result


class WasmApp:
    def __init__(
        self, path_list: typing.List[str], wasm_path: str, should_exit: bool = False
    ):
        import wasmer
        import wasmer_compiler_cranelift

        self.path_list = path_list
        self.wasm_path = wasm_path
        self.should_exit = should_exit

        self.store = wasmer.Store(wasmer.engine.JIT(wasmer_compiler_cranelift.Compiler))
        self.module = wasmer.Module(self.store, open(wasm_path, "rb").read())

        def kclvm_plugin_invoke_json_wasm(method: int, args: int, kwargs: int) -> int:
            return self._invoke_json_wasm(method, args, kwargs)

        self.import_object = wasmer.ImportObject()
        self.import_object.register(
            "env",
            {
                "kclvm_plugin_invoke_json_wasm": wasmer.Function(
                    self.store, kclvm_plugin_invoke_json_wasm
                ),
            },
        )

        self.instance = wasmer.Instance(self.module, self.import_object)
        self.memory = self.instance.exports.memory

        self._invoke_json_buffer = self.instance.exports.kclvm_value_Str(0)
        self.instance.exports.kclvm_value_Str_resize(
            self._invoke_json_buffer, 1024 * 1024 * 1024
        )

    def _invoke_json_wasm(self, method_ptr: int, args_ptr: int, kwargs_ptr: int) -> int:
        method_len = self.instance.exports.kclvm_strlen(method_ptr)
        args_len = self.instance.exports.kclvm_strlen(args_ptr)
        kwargs_len = self.instance.exports.kclvm_strlen(kwargs_ptr)

        reader = bytearray(self.memory.buffer)

        method = reader[method_ptr : method_ptr + method_len].decode()
        args = reader[args_ptr : args_ptr + args_len].decode()
        kwargs = reader[kwargs_ptr : kwargs_ptr + kwargs_len].decode()

        json_result = self._call_py_method(method, args, kwargs)

        bytes_result = json_result.encode(encoding="utf8")
        self.instance.exports.kclvm_value_Str_resize(
            self._invoke_json_buffer, len(bytes_result) + 1
        )

        buf_ptr = self.instance.exports.kclvm_value_Str_ptr(self._invoke_json_buffer)
        buf_len = self.instance.exports.kclvm_value_Str_len(self._invoke_json_buffer)
        assert buf_len == len(bytes_result) + 1

        mem = self.memory.uint8_view()
        for i in range(len(bytes_result)):
            mem[buf_ptr + i] = bytes_result[i]
        mem[buf_ptr + len(bytes_result)] = 0

        return buf_ptr

    def run(self) -> objpkg.KCLResult:
        ctx = self.instance.exports.kclvm_context_new()
        result = self.instance.exports.kclvm_main(ctx)

        c_str_ptr = self.instance.exports.kclvm_value_Str_ptr(result)
        c_str_len = self.instance.exports.kclvm_value_len(result)

        reader = bytearray(self.memory.buffer)
        result = reader[c_str_ptr : c_str_ptr + c_str_len].decode()

        return self.json_to_object(result, None)

    def json_to_object(
        self, json_result: str, warn_json_result: str = None
    ) -> objpkg.KCLResult:
        if kclvm.config.list_option_mode:
            print(json_result, end="")
            return objpkg.KCLResult({})

        data = json.loads(json_result)

        panic_info = {}
        if KCLVM_PANIC_INFO_KEY in data:
            panic_info = data
        else:
            if warn_json_result:
                panic_info = json.loads(warn_json_result)
            else:
                panic_info = {}

        # check panic_info
        if panic_info.get(KCLVM_PANIC_INFO_KEY):
            err_type_code = panic_info["err_type_code"]
            if err_type_code:
                err_type = kcl_error.ErrType((err_type_code,))
            else:
                err_type = kcl_error.ErrType.EvaluationError_TYPE

            file_msg = [
                kcl_error.ErrFileMsg(
                    filename=panic_info.get("kcl_file"),
                    line_no=panic_info.get("kcl_line"),
                    col_no=panic_info.get("kcl_col"),
                    arg_msg=panic_info.get("kcl_arg_msg"),
                )
            ]
            if kclvm.config.debug and kclvm.config.verbose >= 2:
                rust_filename = panic_info.get("rust_file")
                rust_line = panic_info.get("rust_line")
                rust_col = panic_info.get("rust_col")
                print(f"Rust error info: {rust_filename}:{rust_line}:{rust_col}")

            config_meta_file_msg = kcl_error.ErrFileMsg(
                filename=panic_info.get("kcl_config_meta_file"),
                line_no=panic_info.get("kcl_config_meta_line"),
                col_no=panic_info.get("kcl_config_meta_col"),
                arg_msg=panic_info.get("kcl_config_meta_arg_msg"),
            )
            if config_meta_file_msg.arg_msg:
                file_msg.append(config_meta_file_msg)

            if panic_info.get("is_warning") or panic_info.get("is_warnning"):
                kcl_error.report_warning(
                    err_type=err_type, file_msgs=[], arg_msg=panic_info.get("message")
                )
            else:
                kcl_error.report_exception(
                    err_type=err_type,
                    file_msgs=file_msg,
                    arg_msg=panic_info.get("message"),
                )

        return objpkg.KCLResult(data, os.path.abspath(self.path_list[-1]))

    def _call_py_method(self, name: str, args_json: str, kwargs_json: str) -> str:
        try:
            return self._call_py_method_unsafe(name, args_json, kwargs_json)
        except Exception as e:
            return json.dumps({"__kcl_PanicInfo__": f"{e}"})

    def _call_py_method_unsafe(
        self, name: str, args_json: str, kwargs_json: str
    ) -> str:
        dotIdx = name.rfind(".")
        if dotIdx < 0:
            return ""

        modulePath = name[:dotIdx]
        mathodName = name[dotIdx + 1 :]

        plugin_name = modulePath[modulePath.rfind(".") + 1 :]

        module = kcl_plugin.get_plugin(plugin_name)
        mathodFunc = None

        for func_name, func in inspect.getmembers(module):
            if func_name == kcl_info.demangle(mathodName):
                mathodFunc = func
                break

        args = []
        kwargs = {}

        if args_json:
            args = json.loads(args_json)
            if not isinstance(args, list):
                return ""
        if kwargs_json:
            kwargs = json.loads(kwargs_json)
            if not isinstance(kwargs, dict):
                return ""

        result = mathodFunc(*args, **kwargs)
        return json.dumps(result)


def fix_ll_local_name(dst_path, src_path: str):
    replaceArgs_old = []
    replaceArgs_new = []

    for i in range(0, 10):
        replaceArgs_old.append(f"%{i}")
        replaceArgs_new.append(f"%local_{i}")

        replaceArgs_old.append(f"\n{i}")
        replaceArgs_new.append(f"\nlocal_{i}")

    with open(src_path, "r") as file:
        code = file.read()

        if platform.system() == "Windows":
            for i in range(0, len(replaceArgs_old)):
                code = code.replace(replaceArgs_old[i], replaceArgs_new[i], -1)

        with open(dst_path, "w") as f:
            f.write(code)



# Copyright 2021 The KCL Authors. All rights reserved.

import json
import sys
import socket
import subprocess
import typing
import time
import pathlib

from dataclasses import dataclass
from http.server import BaseHTTPRequestHandler, HTTPServer

import kclvm.kcl.ast as ast
import kclvm.kcl.error as kcl_error
import kclvm.kcl.types as types
import kclvm.config
import kclvm.compiler.extension.plugin as plugin
import kclvm.compiler.parser.lark_parser as lark_parser
import kclvm.compiler.parser.parser as parser
import kclvm.program.exec as kclvm_exec
import kclvm.vm.planner as planner
import kclvm.internal.gpyrpc.gpyrpc_pb2 as pb2
import kclvm.internal.gpyrpc.gpyrpc_pb_protorpc as pbrpc
import kclvm.internal.gpyrpc.protorpc as protorpc
from kclvm.tools.format import kcl_fmt_source, kcl_fmt
from kclvm.tools.lint.lint import kcl_lint
from kclvm.tools.query import override_file
from kclvm.tools.validation import validate_code_with_attr_data
from kclvm.tools.langserver import grpc_wrapper
from kclvm.tools.printer import SchemaRuleCodeSnippet, splice_schema_with_rule
from kclvm.tools.list_attribute.schema import get_schema_type_from_code

import kclvm.kcl.error.kcl_err_template as kcl_err_template

import google.protobuf.json_format as _json_format

USAGE = """\
usage: kclvm -m kclvm.program.rpc-server             # run server on stdin/stdout
       kclvm -m kclvm.program.rpc-server -http=:2021 # run fast http server on port
       kclvm -m kclvm.program.rpc-server -h
"""


@dataclass
class CmdFlags:
    help: bool = False

    http_addr: str = ""
    http_port: int = 0

    use_http_server: bool = False


def parse_flags(args: typing.List[str]) -> CmdFlags:
    m = CmdFlags()
    for s in args:
        if s == "-h" or s == "-help":
            m.help = True
            continue

        if s.startswith("-http="):
            value = s[len("-http=") :]
            m.http_addr = str(value[: value.find(":")])
            m.http_port = int(value[value.find(":") + 1 :])
            continue

        if s.startswith("-http.server="):
            value = s[len("-http.server=") :]
            m.http_addr = str(value[: value.find(":")])
            m.http_port = int(value[value.find(":") + 1 :])
            m.use_http_server = True
            continue

    if m.http_port and not m.http_addr:
        m.http_addr = "localhost"

    return m


class ASTTransformer(ast.TreeTransformer):
    def walk_NameConstantLit(self, t: ast.NameConstantLit):
        t.value = str(t.value)
        return t


# KclvmService implementation
class KclvmServiceImpl(pbrpc.KclvmService):
    def Ping(self, args: pb2.Ping_Args) -> pb2.Ping_Result:
        return pb2.Ping_Result(value=args.value)

    def ParseFile_LarkTree(
        self, args: pb2.ParseFile_LarkTree_Args
    ) -> pb2.ParseFile_LarkTree_Result:
        result = pb2.ParseFile_LarkTree_Result()

        if args.source_code:
            t = lark_parser.ParseFile(
                args.filename, args.source_code, ignore_file_line=args.ignore_file_line
            )
            result.lark_tree_json = json.dumps(t, default=lambda obj: obj.__dict__)

        return result

    def ParseFile_AST(self, args: pb2.ParseFile_AST_Args) -> pb2.ParseFile_AST_Result:
        t = parser.ParseFile(args.filename, args.source_code)
        ASTTransformer().walk(t)
        result = pb2.ParseFile_AST_Result()
        result.ast_json = t.to_json()
        return result

    def ParseProgram_AST(
        self, args: pb2.ParseProgram_AST_Args
    ) -> pb2.ParseProgram_AST_Result:
        program = parser.LoadProgram(*list(args.k_filename_list))
        for pkgpath in program.pkgs:
            for i, module in enumerate(program.pkgs[pkgpath]):
                program.pkgs[pkgpath][i] = ASTTransformer().walk(module)
        result = pb2.ParseProgram_AST_Result()
        result.ast_json = program.to_json()
        return result

    def ExecProgram(self, args: pb2.ExecProgram_Args) -> pb2.ExecProgram_Result:
        cmd_args: typing.List[ast.CmdArgSpec] = []
        cmd_overrides: typing.List[ast.CmdOverrideSpec] = []

        # kcl -D name=value main.k
        for x in args.args:
            cmd_args.append(ast.CmdArgSpec(name=x.name, value=x.value))

        # kcl main.k -O pkgpath:path.to.field=field_value
        for x in args.overrides:
            cmd_overrides.append(
                ast.CmdOverrideSpec(
                    pkgpath=x.pkgpath or "__main__",
                    field_path=x.field_path,
                    field_value=x.field_value,
                    action=ast.OverrideAction(x.action)
                    if x.action
                    else ast.OverrideAction.CREATE_OR_UPDATE,
                )
            )

        work_dir: str = args.work_dir
        k_filename_list: typing.List[str] = list(args.k_filename_list)
        k_filename_list = [
            str(
                pathlib.Path(work_dir or kclvm.config.current_path or "").joinpath(file)
            )
            if file.startswith(".")
            else file
            for file in k_filename_list or []
        ]
        k_code_list: typing.List[str] = list(args.k_code_list)
        disable_yaml_result: bool = args.disable_yaml_result
        print_override_ast: bool = args.print_override_ast

        # -r --strict-range-check
        strict_range_check: bool = args.strict_range_check
        # -n --disable-none
        disable_none: bool = args.disable_none
        # -v --verbose
        verbose: int = args.verbose
        # -d --debug
        debug: int = args.debug

        sort_keys: bool = args.sort_keys
        include_schema_type_path: bool = args.include_schema_type_path

        start_time = time.time()
        kcl_result = kclvm_exec.Run(
            k_filename_list,
            work_dir=work_dir,
            k_code_list=k_code_list,
            cmd_args=cmd_args,
            cmd_overrides=cmd_overrides,
            print_override_ast=print_override_ast,
            strict_range_check=strict_range_check,
            disable_none=disable_none,
            verbose=verbose,
            debug=debug,
        )
        end_time = time.time()

        result = pb2.ExecProgram_Result()

        result.escaped_time = f"{end_time-start_time}"

        # json
        output_json = planner.JSONPlanner(
            sort_keys=sort_keys, include_schema_type_path=include_schema_type_path
        ).plan(
            kcl_result.filter_by_path_selector(
                to_kcl=not kclvm.config.is_target_native
            ),
            to_py=not kclvm.config.is_target_native,
        )
        result.json_result = output_json

        # yaml
        if not disable_yaml_result:
            output_yaml = planner.YAMLPlanner(
                sort_keys=sort_keys, include_schema_type_path=include_schema_type_path
            ).plan(
                kcl_result.filter_by_path_selector(
                    to_kcl=not kclvm.config.is_target_native
                ),
                to_py=not kclvm.config.is_target_native,
            )
            result.yaml_result = output_yaml

        return result

    def ResetPlugin(self, args: pb2.ResetPlugin_Args) -> pb2.ResetPlugin_Result:
        plugin.reset_plugin(args.plugin_root)
        result = pb2.ResetPlugin_Result()
        return result

    def FormatCode(self, args: pb2.FormatCode_Args) -> pb2.FormatCode_Result:
        formatted, _ = kcl_fmt_source(args.source)
        return pb2.FormatCode_Result(formatted=formatted.encode("utf-8"))

    def FormatPath(self, args: pb2.FormatPath_Args) -> pb2.FormatPath_Result:
        path = args.path
        recursively = False
        if path.endswith("..."):
            recursively = True
            path = path[: len(path) - 3]
            if path == "" or path is None:
                path = "."
        changed_paths = kcl_fmt(path, recursively=recursively)
        return pb2.FormatPath_Result(changedPaths=changed_paths)

    def LintPath(self, args: pb2.LintPath_Args) -> pb2.LintPath_Result:
        path = args.path

        results: typing.List[str] = []
        for lintMessage in kcl_lint(path):
            results.append(f"{lintMessage.msg}")

        return pb2.LintPath_Result(results=results)

    def OverrideFile(self, args: pb2.OverrideFile_Args) -> pb2.OverrideFile_Result:
        result = override_file(args.file, args.specs, args.import_paths)
        return pb2.OverrideFile_Result(result=result)

    def EvalCode(self, args: pb2.EvalCode_Args) -> pb2.EvalCode_Result:
        import tempfile
        import os

        work_dir = tempfile.mkdtemp()

        with open(f"{work_dir}/kcl.mod", "w") as f:
            pass
        with open(f"{work_dir}/main.k", "w") as f:
            f.write(args.code)

        kcl_result = kclvm_exec.Run(
            ["main.k"], work_dir=work_dir, k_code_list=[args.code]
        )
        output_json = planner.JSONPlanner().plan(
            kcl_result.filter_by_path_selector(), only_first=True
        )

        os.remove(f"{work_dir}/kcl.mod")
        os.remove(f"{work_dir}/main.k")

        result = pb2.EvalCode_Result(json_result=output_json)
        return result

    def ResolveCode(self, args: pb2.ResolveCode_Args) -> pb2.ResolveCode_Result:
        import tempfile
        import os

        work_dir = tempfile.mkdtemp()

        with open(f"{work_dir}/kcl.mod", "w") as f:
            pass
        with open(f"{work_dir}/main.k", "w") as f:
            f.write(args.code)

        ast_prog = parser.LoadProgram(
            *["main.k"],
            work_dir=work_dir,
            k_code_list=[args.code],
        )
        types.ResolveProgram(ast_prog)

        os.remove(f"{work_dir}/kcl.mod")
        os.remove(f"{work_dir}/main.k")

        result = pb2.ResolveCode_Result(success=True)
        return result

    def GetSchemaType(self, args: pb2.GetSchemaType_Args) -> pb2.GetSchemaType_Result:
        schema_type_list = get_schema_type_from_code(
            args.file, args.code, args.schema_name
        )
        return pb2.GetSchemaType_Result(schema_type_list=schema_type_list)

    def ValidateCode(self, args: pb2.ValidateCode_Args) -> pb2.ValidateCode_Result:
        data: str = args.data
        code: str = args.code
        schema: typing.Optional[str] = args.schema or None
        format_: str = args.format or "JSON"

        success = validate_code_with_attr_data(data, code, schema, format_)
        return pb2.ValidateCode_Result(success=success)

    def SpliceCode(self, args: pb2.SpliceCode_Args):
        code_snippets_pb = args.codeSnippets
        code_snippets = [
            SchemaRuleCodeSnippet(
                schema=code_snippet.schema,
                rule=code_snippet.rule,
            )
            for code_snippet in code_snippets_pb
        ]
        splice_code = splice_schema_with_rule(code_snippets)
        return pb2.SpliceCode_Result(spliceCode=splice_code)

    def Complete(self, args: pb2.Complete_Args) -> pb2.Complete_Result:
        pos: pb2.Position = args.pos
        name: str = args.name
        code: str = args.code

        complete_items = grpc_wrapper.complete_wrapper(pos=pos, name=name, code=code)
        return pb2.Complete_Result(completeItems=complete_items)

    def GoToDef(self, args: pb2.GoToDef_Args) -> pb2.GoToDef_Result:
        pos: pb2.Position = args.pos
        code: str = args.code

        locations = grpc_wrapper.go_to_def_wrapper(pos=pos, code=code)
        return pb2.GoToDef_Result(locations=locations)

    def DocumentSymbol(
        self, args: pb2.DocumentSymbol_Args
    ) -> pb2.DocumentSymbol_Result:
        file: str = args.file
        code: str = args.code
        symbol = grpc_wrapper.document_symbol_wrapper(file=file, code=code)
        return pb2.DocumentSymbol_Result(symbol=symbol)

    def Hover(self, args: pb2.Hover_Args) -> pb2.Hover_Result:
        pos: pb2.Position = args.pos
        code: str = args.code
        hover_result = grpc_wrapper.hover_wrapper(pos=pos, code=code)
        return pb2.Hover_Result(hoverResult=hover_result)

    _kcl_go_exe: str = ""

    def ListDepFiles(self, args: pb2.ListDepFiles_Args) -> pb2.ListDepFiles_Result:
        if not self._kcl_go_exe:
            import os

            if os.name == "nt":
                _executable_root = os.path.dirname(sys.executable)
                self._kcl_go_exe = f"{_executable_root}/kcl-go.exe"
            else:
                _executable_root = os.path.dirname(os.path.dirname(sys.executable))
                self._kcl_go_exe = f"{_executable_root}/bin/kcl-go"

        # kcl-go list-app -use-fast-parser=<bool> -show-abs=<bool> -show-index=false <work_dir>
        args = [
            self._kcl_go_exe,
            "list-app",
            f"-use-fast-parser={args.use_fast_parser}",
            f"-include-all={args.include_all}",
            f"-show-abs={args.use_abs_path}",
            "-show-index=false",
            args.work_dir,
        ]

        proc = subprocess.run(args, capture_output=True, text=True)
        stdout = str(proc.stdout or "").strip()
        stderr = str(proc.stderr or "").strip()

        if proc.returncode != 0:
            if stdout and stderr:
                raise Exception(f"stdout: {stdout}, stderr: {stderr}")
            else:
                raise Exception(stdout if stdout else stderr)

        pkgroot: str = ""
        pkgpath: str = ""
        files: typing.List[str] = []

        for s in stdout.splitlines():
            if s.startswith("pkgroot:"):
                pkgroot = s[len("pkgroot:") :]
                pkgroot = s.strip()
            elif s.startswith("pkgpath:"):
                pkgpath = s[len("pkgpath:") :]
                pkgpath = s.strip()
            else:
                s = s.strip()
                if s:
                    files.append(s)

        return pb2.ListDepFiles_Result(pkgroot=pkgroot, pkgpath=pkgpath, files=files)

    def LoadSettingsFiles(
        self, args: pb2.LoadSettingsFiles_Args
    ) -> pb2.LoadSettingsFiles_Result:
        return kclvm.config.load_settings_files(args.work_dir, args.files)


def _makeRpcServer():
    rpc_server = protorpc.Server()

    # BuiltinService implementation (depends on rpc_server)
    class BuiltinServiceImpl(pbrpc.BuiltinService):
        def Ping(self, args: pb2.Ping_Args) -> pb2.Ping_Result:
            return pb2.Ping_Result(value=args.value)

        def ListMethod(self, args: pb2.ListMethod_Args) -> pb2.ListMethod_Result:
            return pb2.ListMethod_Result(
                method_name_list=rpc_server.get_method_name_list()
            )

    # Metaclass for packaging services
    builtin_service = pbrpc.BuiltinService_Meta(
        typing.cast(pbrpc.BuiltinService, BuiltinServiceImpl())
    )
    kclvm_service = pbrpc.KclvmService_Meta(
        typing.cast(pbrpc.KclvmService, KclvmServiceImpl())
    )

    # Register service metaclass
    rpc_server.register_service(builtin_service)
    rpc_server.register_service(kclvm_service)

    return rpc_server


def runStdioProtorpcServer():
    """Start protorpc service based on stdin/stdout"""

    # Redirect stdout: raw_stdout raw_stdout will be used for protorpc communication
    raw_stdout = sys.stdout
    sys.stdout = sys.stderr

    # Start the service based on raw_stdout (blocking)
    rpc_server = _makeRpcServer()
    rpc_server.run(stdout=raw_stdout)


def runHttpServer(*, addr: str = "", port: int = 2021):

    rpc_server = _makeRpcServer()

    class MyHTTPServer(HTTPServer):
        def server_bind(self):
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            self.socket.bind(self.server_address)

    class httpHandler(BaseHTTPRequestHandler):
        def do_GET(self):

            # http://localhost:2021/api:protorpc/BuiltinService.Ping
            # http://localhost:2021/api:protorpc/BuiltinService.ListMethod

            # Build protorpc call
            if self.path.startswith("/api:protorpc/"):
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()

                from urllib.parse import urlparse

                u = urlparse(self.path)
                method = u.path[len("/api:protorpc/") :]

                resp = {}
                try:
                    result = rpc_server.call_method(method, b"{}", encoding="json")
                    resp["result"] = result
                    resp["error"] = ""
                except kcl_error.KCLException as err:
                    resp["result"] = None
                    resp["error"] = f"{err}"
                    resp["kcl_err"] = kcl_err_to_response_dict(err)
                except Exception as err:
                    resp["result"] = None
                    resp["error"] = f"{err}"
                self.wfile.write(bytes(json.dumps(resp), "utf8"))
            else:
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()

                message = f"Hello, World! Here is a GET response: path={self.path}"
                self.wfile.write(bytes(message, "utf8"))

        def do_POST(self):
            self.send_response(200)
            self.send_header("Content-type", "application/octet-stream")
            self.end_headers()

            content_len = int(self.headers.get("Content-Length"))
            req_body = self.rfile.read(content_len)

            # Build protorpc call
            if self.path.startswith("/api:protorpc/"):
                method = self.path[len("/api:protorpc/") :]
                resp = {}
                try:
                    result = rpc_server.call_method(method, req_body, encoding="json")
                    resp["result"] = result
                    resp["error"] = ""
                except kcl_error.KCLException as err:
                    resp["result"] = None
                    resp["error"] = f"{err}"
                    resp["kcl_err"] = kcl_err_to_response_dict(err)
                except Exception as err:
                    resp["result"] = None
                    resp["error"] = f"{err}"
                self.wfile.write(bytes(json.dumps(resp), "utf8"))
            else:
                message = f"Hello, World! Here is a POST response: path={self.path}"
                self.wfile.write(bytes(message, "utf8"))

    print(f"run http server on http://{addr}:{port} ...")

    with MyHTTPServer(("", port), httpHandler) as server:
        server.serve_forever()


def runFastApiServer(*, addr: str = "", port: int = 2021):
    import uvicorn

    app = create_app()
    uvicorn.run(app, host=addr, port=port)


def kcl_err_to_response_dict(err: kcl_error.KCLException) -> pb2.KclError:
    return _json_format.MessageToDict(
        pb2.KclError(
            ewcode=f"{err.ewcode}",
            name=f"{err.name}",
            msg=f"{err.arg_msg}",
            error_infos=[
                pb2.KclErrorInfo(
                    err_level=f"{err_info.err_level}",
                    arg_msg=f"{err_info.arg_msg}",
                    filename=f"{err_info.filename}",
                    src_code=f"{kcl_err_template.get_src_code(err_info)}",
                    line_no=f"{err_info.line_no}",
                    col_no=f"{err_info.col_no}",
                )
                for err_info in err.err_info_stack
            ],
        ),
        including_default_value_fields=True,
        preserving_proto_field_name=True,
    )


# ./gunicorn "kclvm.program.rpc-server:create_app()" -w 4 -k uvicorn.workers.UvicornWorker -b :2021
def create_app():
    import fastapi

    app = fastapi.FastAPI()
    rpc_server = _makeRpcServer()

    @app.get("/")
    async def index():
        return "KCL Rest Server"

    @app.get("/api:protorpc/{method}")
    async def on_rest_api_get(method: str, request: fastapi.Request):
        resp = {}
        try:
            result = rpc_server.call_method(method, b"{}", encoding="json")
            resp["result"] = result
            resp["error"] = ""
        except kcl_error.KCLException as err:
            resp["result"] = None
            resp["error"] = f"{err}"
            resp["kcl_err"] = kcl_err_to_response_dict(err)
        except Exception as err:
            resp["result"] = None
            resp["error"] = f"{err}"

        return resp

    @app.post("/api:protorpc/{method}")
    async def on_rest_api_post(method: str, request: bytes = fastapi.Body(...)):
        resp = {}
        try:
            result = rpc_server.call_method(method, request, encoding="json")
            resp["result"] = result
            resp["error"] = ""
        except kcl_error.KCLException as err:
            resp["result"] = None
            resp["error"] = f"{err}"
            resp["kcl_err"] = kcl_err_to_response_dict(err)
        except Exception as err:
            resp["result"] = None
            resp["error"] = f"{err}"

        return resp

    return app


def main():
    flags = parse_flags(sys.argv[1:])

    if flags.help:
        print(USAGE)
        sys.exit(0)

    from kclvm.compiler.parser.lark_parser import GetKclLarkParser

    GetKclLarkParser()

    if flags.http_addr or flags.http_port:
        print(f"flags: {flags}")
        if not flags.use_http_server:
            runFastApiServer(addr=flags.http_addr, port=flags.http_port)
        else:
            runHttpServer(addr=flags.http_addr, port=flags.http_port)
        sys.exit(0)
    else:
        runStdioProtorpcServer()
        sys.exit(0)


if __name__ == "__main__":
    main()



import json
from typing import Union, Optional

from kclvm.internal.gpyrpc.gpyrpc_pb2 import Position
import kclvm.kcl.ast as ast
from .complete import complete
from .go_to_def import go_to_def
from .document_symbol import document_symbol
from .hover import hover


def go_to_def_wrapper(pos: Position, code: str = None) -> str:
    pos_wrapper = proto_pos_to_ast_pos(pos)
    result = go_to_def(pos=pos_wrapper, code=code)
    return json.dumps(obj=result, default=lambda x: x.__dict__)


def complete_wrapper(pos: Position, name: str, code: str = None) -> str:
    pos_wrapper = proto_pos_to_ast_pos(pos)
    result = complete(pos=pos_wrapper, name=name, code=code)
    return json.dumps(obj=result, default=lambda x: x.__dict__)


def proto_pos_to_ast_pos(pos: Position) -> ast.Position:
    return ast.Position(filename=pos.filename, line=pos.line + 1, column=pos.column + 1)


class SnakeToCamel(json.JSONEncoder):
    """Class attributes need to be converted to camel-case notation because client is expecting that."""

    def to_camel(self, s: str):
        return "".join(
            word.capitalize() if idx > 0 else word
            for idx, word in enumerate(s.split("_"))
        )

    def convert_key(
        self, instance: Union[int, float, str, bool, list, dict, tuple]
    ) -> Optional[Union[int, float, str, bool, list, dict, tuple]]:
        if instance is None:
            return None
        if isinstance(instance, (bool, int, float, str)):
            return instance
        elif isinstance(instance, (list, set, tuple)):
            return [self.convert_key(v) for v in instance]
        elif isinstance(instance, dict):
            return {self.to_camel(k): self.convert_key(v) for k, v in instance.items()}
        else:
            return self.convert_key(instance.__dict__)

    def default(self, obj):
        return self.convert_key(obj.__dict__)


def document_symbol_wrapper(file: str, code: str = None) -> str:
    result = document_symbol(file=file, code=code)
    return json.dumps(obj=result, cls=SnakeToCamel)


def hover_wrapper(pos: Position, code: str) -> str:
    pos_wrapper = proto_pos_to_ast_pos(pos)
    result = hover(pos=pos_wrapper, code=code)
    return json.dumps(obj=result, default=lambda x: x.__dict__)
